# This work is based upon CMFExternalFile, which was developed by
# Alan Runyan with contributions by Andy McKay, Kiran Jonnalagadda
# and PloneExternalFile developed by Florian Geier.
#
# CMFManagedFile extends CMFExternalFile to include a concept that
# can be found in PloneExternalFile where uploaded documents may be
# routed to specific directories or file systems.  This is specified by the
# user at upload time through the selection of pre-defined "repositories"
# that the site administrator would define and establish.
#
# The merger of CMFExternalFile and PloneExternalFile has been further
# extended to include some aspects of managing the external files on the
# file systems.  These include the automatic deletion of the file system files
# for deleted documents of the class CMFManagedFile governed by customizable
# and user selectable deletion policies, and the automatic movement of files
# from one file system repository to another based upon an edit action
# that includes the selection of a repository other than the repository
# in which the file is currently stored.
#
"""
$Id: content.py 992 2005-02-10 02:55:45Z jlaplace $
"""
import os, sys

if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

import unittest
from Testing import ZopeTestCase
from Products.CMFManagedFile.utils import CallbackStreamWrapper, LStringIO
from Products.CMFManagedFile.utils import LimitedReadCallback

class CallbackTester:
    def __init__(self, testcase):
        self.testcase = testcase
        self.reset()

    def reset(self):
        self.steps = []
        self.last_total = None
        self.last_position = None

    start = exception = reset

    def __call__(self, total, position):
        if self.last_total is None:
            self.last_total = total
        else:
            self.testcase.failUnlessEqual(total, self.last_total)
        if self.last_position is None:
            self.last_position = position
        else:
            if position < self.last_position:
                self.testcase.fail("Last pos was %d, but this is %d" \
                                   % (self.last_position, position))
        if position > total:
            self.testcase.fail("Total is %d, but current pos is greater: %d" \
                               % (total, position))
        self.steps.append(total / 100 * position)
        return True

# simulates a socket for exercising our caching and chunked reads.
class SlowReader:
    def __init__(self, testcase, data):
        self.testcase = testcase
        self.data = data
        self.pos = 0

    def __len__(self):
        return len(self.data)

    def read(self, n):
        # always give less than they want, but no less than 32.
        n = max(32, n/2)
        if self.pos >= len(self.data):
            self.data = "" # to simulate a consumed hacked socket.
            return ''
        data = self.data[self.pos:self.pos+n]
        self.pos += n
        return data

class StreamWrapperTest(unittest.TestCase):

    def setUp(self):
        self.data = '1234567890' * 1024 * 2
        self.stream = LStringIO(self.data)
        self.cb = CallbackTester(self)

    def tearDown(self):
        self.cb = None

    def get_stream(self, cb=None):
        if cb is None:
            cb = self.cb
        return CallbackStreamWrapper(self.stream, cb, 1024)

    def test_unbound_read(self):
        s = self.get_stream()
        self.assertEquals(self.data, s.read())
        self.assertEquals(len(self.cb.steps), 20)

    def test_unbound_double_read(self):
        s = self.get_stream()
        self.assertEquals(self.data, s.read())
        self.assertEquals(len(self.cb.steps), 20)
        s.seek(0)
        self.assertEquals(self.data, s.read())
        self.assertEquals(len(self.cb.steps), 20)

    def test_bound_read(self):
        s = self.get_stream()
        self.assertEquals(self.data[:160], s.read(160))
        self.assertEquals(len(self.cb.steps), 1)

    def test_double_bound_read_then_unbound(self):
        s = self.get_stream()
        self.assertEquals(self.data[:111], s.read(111))
        self.assertEquals(len(self.cb.steps), 1)
        s.read(3)
        self.assertEquals(self.data[114:225], s.read(111))
        self.assertEquals(len(self.cb.steps), 1)
        s.seek(0)
        self.assertEquals(self.data[:116], s.read(116))
        s.seek(0)
        s.read(10)
        self.assertEquals(self.data[10:222], s.read(212))
        self.assertEquals(self.data[222:], s.read())
        self.assertEquals(len(self.cb.steps), 21) # ???
        s.seek(2048)
        self.assertEquals(self.data[2048:], s.read())
        self.assertEquals(len(self.cb.steps), 18)

    def test_cache_chunked(self):
        stream = SlowReader(self, self.data)
        s = CallbackStreamWrapper(stream, self.cb, 1024)
        chunks = []
        while 1:
            chunk = s.read(128)
            if not chunk:
                break
            chunks.append(chunk)
        self.failUnlessEqual(len(''.join(chunks)), len(self.data))
        # Now do it again - the cache will kick in.
        # The underlying stream should be reporting zero to trigger the cache
        self.failIf(len(stream), len(stream))
        s.seek(0)
        chunks = []
        while 1:
            chunk = s.read(128)
            if not chunk:
                break
            chunks.append(chunk)
        self.failUnlessEqual(len(''.join(chunks)), len(self.data))

    def test_cache_block(self):
        # And reading as a chunk
        stream = SlowReader(self, self.data)
        s = CallbackStreamWrapper(stream, self.cb, 1024)
        got = s.read()
        self.failUnlessEqual(len(got), len(self.data))
        # Now do it again - the cache will kick in.
        s.seek(0)
        got = s.read()
        self.failUnlessEqual(len(got), len(self.data))

    def test_limiter_below_bound(self):
        cb = LimitedReadCallback(len(self.data) + 1)
        s = self.get_stream(cb=cb)
        # Should be able to read the whole stream without problem.
        got = s.read()
        self.failUnlessEqual(len(got), len(self.data))
        # Now do it again - the cache will kick in.
        s.seek(0)
        got = s.read()
        self.failUnlessEqual(len(got), len(self.data))

    def test_limiter_above_bound(self):
        cb = LimitedReadCallback(1 << 10)
        s = self.get_stream(cb=cb)
        # Should be able to read half the limit without problem.
        s.read(1 << 9)
        # Now read another half
        s.read(1 << 9)
        # And now read a byte, which should hit the limit
        self.assertRaises(ValueError, s.read, 1)

def test_suite():
    t = unittest.TestSuite()
    t.addTest(unittest.makeSuite(StreamWrapperTest))
    return t

if __name__ == '__main__':
    unittest.main()
